/*

    File: bindings_ARB_shader_objects.cpp

    Copyright (C) 2000-2010 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <iostream>
#include "system/gl.h"
#include "base/opengl.h"
#include "shader/shaderext.h"
#include "program.h"
#include "bindings_ARB_shader_objects.h"

using namespace std;

LUAF2glh	(glAttachObjectARB)
LUAFglh		(glCompileShaderARB)
LUAFrglh	(glCreateProgramObjectARB)
LUAFirglh	(glCreateShaderObjectARB)
LUAFglh		(glDeleteObjectARB)
LUAF2glh	(glDetachObjectARB)
LUAFirglh	(glGetHandleARB)

LUAFstart(glGetInfoLogARB)
	GLhandleARB handle = LUAPglh(1);

	GLint infoLogLength = 0;
	glGetObjectParameterivARB(handle, GL_OBJECT_INFO_LOG_LENGTH_ARB, &infoLogLength);
	if (!infoLogLength) {
		lua_pushnil(L);
		return 1;
	}
	
	char *infoLog = new char[infoLogLength];
	GLsizei charsWritten = 0;
	glGetInfoLogARB(handle, infoLogLength, &charsWritten, infoLog);
	lua_pushstring(L, infoLog);
	delete infoLog;
	return 1;
LUAFend()

//I'm guessing certain params have different types and lengths
//but all I see would imply a single int ...
LUAFstart(glGetObjectParameterivARB)
	GLint ret = 0;
	glGetObjectParameterivARB(LUAPglh(1), LUAPi(2), &ret);
	lua_pushinteger(L, ret);
	return 1;
LUAFend()

//just like glGetAttribLocationARB
LUAFstart(glGetUniformLocationARB)
	GLhandleARB progh = (GLhandleARB)(int)lua_tonumber(L,1);
	if (!lua_isstring(L,2)) {
		con << "glGetUniformLocation expected a string for its second param" << endl;
		return 0;
	}
	lua_pushinteger(L, glGetUniformLocationARB(progh, lua_tostring(L,2)));
	return 1;
LUAFend()

LUAFglh(glLinkProgramARB)

LUAFstart(glShaderSourceARB)
	const char *str = lua_tostring(L, 2);
	glShaderSourceARB(LUAPglh(1), 1, &str, NULL);
LUAFend()

/*
if there was a type implicit version of glUniform, how should the type be deduced?
we have at our disposal (a) the uniform type and (b) the lua passed type ...
chances are it'll be solely derived from the uniform type

so the general handles glUniformi, glUniformf, and glUniformMatrix (which is all floats)
it'll handle any type
win.
*/
LUAFstart(glUniform1fARB)
	if (argc < 1) {
		con << "glUniform expect at least one param" << endl;
		return 0;
	}
	
	GLhandleARB handle = glGetHandleARB(GL_PROGRAM_OBJECT_ARB);
	int loc = -1;
	if (lua_isnumber(L,1)) {
		loc = (int)lua_tonumber(L,1);
	} else if (lua_isstring(L,1)) {
		loc = glGetUniformLocationARB(handle, lua_tostring(L,1));
	}
	
	GLsizei length = 0;
	GLsizei size = 0;
	GLenum type = 0;
	char name[256];
	glGetActiveUniformARB(handle, loc, sizeof(name), &length, &size, &type, name);
	
	switch (type) {
	case GL_FLOAT:
	case GL_FLOAT_VEC2:
	case GL_FLOAT_VEC3:
	case GL_FLOAT_VEC4:
		if (argc == 2) {
			if (lua_istable(L,2)) {
				int len = lua_objlen(L,2);
				if (size == 1) {	//shortcut for non-array variables
					switch (len) {
					case 1: glUniform1fvARB(loc, 1, luaReadVec1f(L) ); break;
					case 2: glUniform2fvARB(loc, 1, luaReadVec2f(L) ); break;
					case 3: glUniform3fvARB(loc, 1, luaReadVec3f(L) ); break;
					case 4: glUniform4fvARB(loc, 1, luaReadVec4f(L) ); break;
					default: con << "glUniformf table parameter got an unknown size " << len << endl; break;
					}
				} else {
					vecf v;
					switch (type) {
					case GL_FLOAT:		luaReadVecDyn<float>(L, v); v.resize(size);		glUniform1fvARB(loc, size, v); break;
					case GL_FLOAT_VEC2: luaReadVecDyn<float>(L, v); v.resize(2 * size);	glUniform2fvARB(loc, size, v); break;
					case GL_FLOAT_VEC3: luaReadVecDyn<float>(L, v); v.resize(3 * size);	glUniform2fvARB(loc, size, v); break;
					case GL_FLOAT_VEC4: luaReadVecDyn<float>(L, v); v.resize(4 * size);	glUniform2fvARB(loc, size, v); break;
					default: assert(false);					
					}
				}
			} else {
				glUniform1fARB(loc, LUAPf(2));
			}
		} else
		LUAFvarcase(3, glUniform2fARB(loc, LUAPf(2), LUAPf(3)))
		LUAFvarcase(4, glUniform3fARB(loc, LUAPf(2), LUAPf(3), LUAPf(4)))
		LUAFvarcase(5, glUniform4fARB(loc, LUAPf(2), LUAPf(3), LUAPf(4), LUAPf(5)))
		LUAFvarendcase("table, 2, 3, 4, or 5")
		break;
		
	case GL_FLOAT_MAT2:
	case GL_FLOAT_MAT3:
	case GL_FLOAT_MAT4:
		if (argc == 2 && lua_istable(L,2)) {
			int len = lua_objlen(L,2);
			if (size == 1) {
				switch (len) {
				case 4: glUniformMatrix2fvARB(loc, 1, GL_FALSE, luaReadVec<4,float>(L) ); break;
				case 9: glUniformMatrix3fvARB(loc, 1, GL_FALSE, luaReadVec<9,float>(L) ); break;
				case 16: glUniformMatrix4fvARB(loc, 1, GL_FALSE, luaReadVec<16,float>(L) ); break;
				default: con << "glUniformMatrix table parameter got an unknown size " << len << endl; break;
				}
			} else {
				vecf v;
				switch (type) {
				case GL_FLOAT_MAT2:	luaReadVecDyn<float>(L, v); v.resize(4 * size); glUniformMatrix2fvARB(loc, size, GL_FALSE, v); break;
				case GL_FLOAT_MAT3:	luaReadVecDyn<float>(L, v); v.resize(9 * size); glUniformMatrix3fvARB(loc, size, GL_FALSE, v); break;
				case GL_FLOAT_MAT4: luaReadVecDyn<float>(L, v); v.resize(16* size); glUniformMatrix4fvARB(loc, size, GL_FALSE, v); break;
				default: assert(false);
				}
			}
		} else
		LUAFvarendcase("uniform, table")
		break;
		
	case GL_INT:
	case GL_INT_VEC2:
	case GL_INT_VEC3:
	case GL_INT_VEC4:
	case GL_BOOL:
	case GL_BOOL_VEC2:
	case GL_BOOL_VEC3:
	case GL_BOOL_VEC4:
	case GL_SAMPLER_1D:
	case GL_SAMPLER_2D:
	case GL_SAMPLER_3D:
	case GL_SAMPLER_CUBE:
	case GL_SAMPLER_1D_SHADOW:
	case GL_SAMPLER_2D_SHADOW:
		if (argc == 2) {
			if (lua_istable(L,2)) {
				int len = lua_objlen(L,2);
				if (size == 1) {
					switch (len) {
					case 1: glUniform1ivARB(loc, 1, luaReadVec<2,GLint>(L) ); break;
					case 2: glUniform2ivARB(loc, 1, luaReadVec<2,GLint>(L) ); break;
					case 3: glUniform3ivARB(loc, 1, luaReadVec<3,GLint>(L) ); break;
					case 4: glUniform4ivARB(loc, 1, luaReadVec<4,GLint>(L) ); break;
					default: con << "glUniform table parameter got an unknown size " << len << endl; break;
					}
				} else {
					_vec<0,GLint> v;
					switch (type) {
					case GL_INT:
					case GL_BOOL:
					case GL_SAMPLER_1D:
					case GL_SAMPLER_2D:
					case GL_SAMPLER_3D:
					case GL_SAMPLER_CUBE:
					case GL_SAMPLER_1D_SHADOW:
					case GL_SAMPLER_2D_SHADOW:
						luaReadVecDyn<GLint>(L, v);
						v.resize(size);
						glUniform1ivARB(loc, size, v);
						break;
					case GL_INT_VEC2:
					case GL_BOOL_VEC2:
						luaReadVecDyn<GLint>(L, v);
						v.resize(2 * size);
						glUniform2ivARB(loc, size, v);
						break;
					case GL_INT_VEC3:
					case GL_BOOL_VEC3:
						luaReadVecDyn<GLint>(L, v);
						v.resize(3 * size);
						glUniform3ivARB(loc, size, v);
						break;
					case GL_INT_VEC4:
					case GL_BOOL_VEC4:
						luaReadVecDyn<GLint>(L, v);
						v.resize(4 * size);
						glUniform4ivARB(loc, size, v);
						break;
					default:
						assert(false);
					}
				}
			} else {
				glUniform1iARB(loc, LUAPf(2));
			}
		} else
		LUAFvarcase(3, glUniform2iARB(loc, LUAPi(2), LUAPi(3)))
		LUAFvarcase(4, glUniform3iARB(loc, LUAPi(2), LUAPi(3), LUAPi(4)))
		LUAFvarcase(5, glUniform4iARB(loc, LUAPi(2), LUAPi(3), LUAPi(4), LUAPi(5)))
		LUAFvarendcase("table, 2, 3, 4, or 5")
		break;
	}
LUAFend()

//I did a macro to autocheck for the gl name of the LUAF func, thats why the names are weird
LUAFstart(glUniform1fvARB)
	if (argc < 1) {
		con << "glUniformf expect at least one param" << endl;
		return 0;
	}
	int loc = -1;
	if (lua_isnumber(L,1)) {
		loc = (int)lua_tonumber(L,1);
	} else if (lua_isstring(L,1)) {
		loc = glGetUniformLocationARB(glGetHandleARB(GL_PROGRAM_OBJECT_ARB), lua_tostring(L,1));
	}
	if (argc == 2) {
		if (lua_istable(L,2)) {
			int len = lua_objlen(L,2);
			if (len == 1) glUniform1fvARB(loc, 1, luaReadVec1f(L) );
			else if (len == 2) glUniform2fvARB(loc, 1, luaReadVec2f(L) );
			else if (len == 3) glUniform3fvARB(loc, 1, luaReadVec3f(L) );
			else if (len == 4) glUniform4fvARB(loc, 1, luaReadVec4f(L) );
			else con << "glUniformf table parameter got an unknown size " << len << endl;
		} else {
			glUniform1fARB(loc, LUAPf(2));
		}
	} else
	LUAFvarcase(3, glUniform2fARB(loc, LUAPf(2), LUAPf(3)))
	LUAFvarcase(4, glUniform3fARB(loc, LUAPf(2), LUAPf(3), LUAPf(4)))
	LUAFvarcase(5, glUniform4fARB(loc, LUAPf(2), LUAPf(3), LUAPf(4), LUAPf(5)))
	LUAFvarendcase("table, 2, 3, 4, or 5")
LUAFend()

LUAFstart(glUniform1ivARB)
	if (argc < 1) {
		con << "glUniformi expect at least one param" << endl;
		return 0;
	}
	int loc = -1;
	if (lua_isnumber(L,1)) {
		loc = (int)lua_tonumber(L,1);
	} else if (lua_isstring(L,1)) {
		loc = glGetUniformLocationARB(glGetHandleARB(GL_PROGRAM_OBJECT_ARB), lua_tostring(L,1));
	}
	if (argc == 2) {
		if (lua_istable(L,2)) {
			int len = lua_objlen(L,2);
			if (len == 1) glUniform1ivARB(loc, 1, luaReadVec<2,GLint>(L) );
			else if (len == 2) glUniform2ivARB(loc, 1, luaReadVec<2,GLint>(L) );
			else if (len == 3) glUniform3ivARB(loc, 1, luaReadVec<2,GLint>(L) );
			else if (len == 4) glUniform4ivARB(loc, 1, luaReadVec<2,GLint>(L) );
			else con << "glUniformi table parameter got an unknown size " << len << endl;
		} else {
			glUniform1iARB(loc, LUAPf(2));
		}
	} else
	LUAFvarcase(3, glUniform2iARB(loc, LUAPi(2), LUAPi(3)))
	LUAFvarcase(4, glUniform3iARB(loc, LUAPi(2), LUAPi(3), LUAPi(4)))
	LUAFvarcase(5, glUniform4iARB(loc, LUAPi(2), LUAPi(3), LUAPi(4), LUAPi(5)))
	LUAFvarendcase("table, 2, 3, 4, or 5")
LUAFend()

LUAFstart(glUniformMatrix2fvARB)
	if (argc < 1) {
		con << "glUniformMatrix expect at least one param" << endl;
		return 0;
	}
	int loc = -1;
	if (lua_isnumber(L,1)) {
		loc = (int)lua_tonumber(L,1);
	} else if (lua_isstring(L,1)) {
		loc = glGetUniformLocationARB(glGetHandleARB(GL_PROGRAM_OBJECT_ARB), lua_tostring(L,1));
	}
	if (argc == 2 && lua_istable(L,2)) {
		int len = lua_objlen(L,2);
		if (len == 4) glUniformMatrix2fvARB(loc, 1, GL_FALSE, luaReadVec<4,float>(L) );
		else if (len == 9) glUniformMatrix3fvARB(loc, 1, GL_FALSE, luaReadVec<9,float>(L) );
		else if (len == 16) glUniformMatrix4fvARB(loc, 1, GL_FALSE, luaReadVec<16,float>(L) );
		else con << "glUniformMatrix table parameter got an unknown size " << len << endl;
	} else
	LUAFvarendcase("uniform, table")
LUAFend()

LUAFglh			(glUseProgramObjectARB)
LUAFglh			(glValidateProgramARB)

void luaBindGL_ARB_shader_objects(lua_State *L) {
	lua_getglobal(L, "gl");

	LUA_GLUE_FUNC_SAFE_ARB(glAttachObject)
	LUA_GLUE_FUNC_SAFE_ARB(glCompileShader)
	LUA_GLUE_FUNC_SAFE_ARB(glCreateProgramObject)
	LUA_GLUE_FUNC_SAFE_ARB(glCreateShaderObject)
	LUA_GLUE_FUNC_SAFE_ARB(glDeleteObject)
	LUA_GLUE_FUNC_SAFE_ARB(glDetachObject)
	LUA_GLUE_FUNC_SAFE_ARB(glGetHandle)
	LUA_GLUE_FUNC_SAFE_ARB(glGetInfoLog)
	//extern void glGetActiveUniformARB(GLhandleARB programObj, GLuint index, GLsizei maxLength, GLsizei *length, GLint *size, GLenum *type, GLcharARB *name);
	//extern void glGetAttachedObjectsARB(GLhandleARB containerObj, GLsizei maxCount, GLsizei *count, GLhandleARB *obj);
	LUA_GLUE_FUNC_SAFE_SUFFIX(glGetObjectParameter, ivARB)
	//extern void glGetShaderSourceARB(GLhandleARB obj, GLsizei maxLength, GLsizei *length, GLcharARB *source);
	//extern void glGetUniformfvARB(GLhandleARB programObj, GLint location, GLfloat *params);
	//extern void glGetUniformivARB(GLhandleARB programObj, GLint location, GLint *params);
	LUA_GLUE_FUNC_SAFE_ARB(glGetUniformLocation)
	LUA_GLUE_FUNC_SAFE_ARB(glLinkProgram)
	LUA_GLUE_FUNC_SAFE_ARB(glShaderSource)
	LUA_GLUE_FUNC_SAFE_SUFFIX(glUniform, 1fARB)
	LUA_GLUE_FUNC_SAFE_GENERAL(glUniformf, glUniform1fvARB)
	LUA_GLUE_FUNC_SAFE_GENERAL(glUniformi, glUniform1ivARB)
	LUA_GLUE_FUNC_SAFE_SUFFIX(glUniformMatrix, 2fvARB)
	LUA_GLUE_FUNC_SAFE_ARB(glUseProgramObject)
	LUA_GLUE_FUNC_SAFE_ARB(glValidateProgram)

	LUAGiARB(GL_PROGRAM_OBJECT)
	LUAGiARB(GL_OBJECT_TYPE)
	LUAGiARB(GL_OBJECT_SUBTYPE)
	LUAGiARB(GL_OBJECT_DELETE_STATUS)
	LUAGiARB(GL_OBJECT_COMPILE_STATUS)
	LUAGiARB(GL_OBJECT_LINK_STATUS)
	LUAGiARB(GL_OBJECT_VALIDATE_STATUS)
	LUAGiARB(GL_OBJECT_INFO_LOG_LENGTH)
	LUAGiARB(GL_OBJECT_ATTACHED_OBJECTS)
	LUAGiARB(GL_OBJECT_ACTIVE_UNIFORMS)
	LUAGiARB(GL_OBJECT_ACTIVE_UNIFORM_MAX_LENGTH)
	LUAGiARB(GL_OBJECT_SHADER_SOURCE_LENGTH)
	LUAGiARB(GL_SHADER_OBJECT)
	LUAGiARB(GL_FLOAT_VEC2)
	LUAGiARB(GL_FLOAT_VEC3)
	LUAGiARB(GL_FLOAT_VEC4)
	LUAGiARB(GL_INT_VEC2)
	LUAGiARB(GL_INT_VEC3)
	LUAGiARB(GL_INT_VEC4)
	LUAGiARB(GL_BOOL)
	LUAGiARB(GL_BOOL_VEC2)
	LUAGiARB(GL_BOOL_VEC3)
	LUAGiARB(GL_BOOL_VEC4)
	LUAGiARB(GL_FLOAT_MAT2)
	LUAGiARB(GL_FLOAT_MAT3)
	LUAGiARB(GL_FLOAT_MAT4)
	LUAGiARB(GL_SAMPLER_1D)
	LUAGiARB(GL_SAMPLER_2D)
	LUAGiARB(GL_SAMPLER_3D)
	LUAGiARB(GL_SAMPLER_CUBE)
	LUAGiARB(GL_SAMPLER_1D_SHADOW)
	LUAGiARB(GL_SAMPLER_2D_SHADOW)
	LUAGiARB(GL_SAMPLER_2D_RECT)
	LUAGiARB(GL_SAMPLER_2D_RECT_SHADOW)
	lua_pop(L,1);
}

/*

and here's someone using them:

vertexCode = [[
varying vec2 pos;
void main() {
	pos.xy = gl_Vertex.xy;
	gl_Position = ftransform();
}
]];
fragmentCode = [[
varying vec2 pos;
uniform sampler2D tex;
void main() {
	gl_FragColor = texture2D(tex, pos);
}
]];

vs = VertexShader.new()
vs:setSource(vertexCode)
vs:compile()
fs = FragmentShader.new()
fs:setSource(fragmentCode)
fs:compile()
sp = ShaderProgram.new()
sp:attach(vs)
sp:attach(fs)
sp:link()
sp:setUniform('tex', 0)

... or, more concisely

vs = VertexShader.new{
	source=vertexCode;
}
fs = FragmentShader.new{
	source=fragmentCode;
}
sp = ShaderProgram.new{
	attach={vs, fs};
	uniforms={tex=0;};
}

... or, more concisely, 

ShaderProgram.new{
	attach={
		VertexShader.new{source=vertexCode};
		FragmentShader.new{source=fragmentCode};
	};
	uniforms={tex=0;};
}

... or, more concisely

ShaderProgram.new{
	vertexCode=vertexCode;
	fragmentCode=fragmentCode;
	uniforms={tex=0;};
}

*/