/*

    File: bindings_ARB_vertex_buffer_object.cpp

    Copyright (C) 2000-2010 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <iostream>
#include "system/gl.h"
#include "system/exception.h"
#include "base/opengl.h"
#include "shader/shaderext.h"
#include "program.h"
#include "bindings_ARB_vertex_buffer_object.h"

using namespace std;

/*
  void BindBufferARB(enum target, uint buffer);
    void DeleteBuffersARB(sizei n, const uint *buffers);
    void GenBuffersARB(sizei n, uint *buffers);
    boolean IsBufferARB(uint buffer);

    void BufferDataARB(enum target, sizeiptrARB size, const void *data,
                       enum usage);
    void BufferSubDataARB(enum target, intptrARB offset, sizeiptrARB size,
                          const void *data);
    void GetBufferSubDataARB(enum target, intptrARB offset,
                             sizeiptrARB size, void *data);

    void *MapBufferARB(enum target, enum access);
    boolean UnmapBufferARB(enum target);

    void GetBufferParameterivARB(enum target, enum pname, int *params);
    void GetBufferPointervARB(enum target, enum pname, void **params);
*/

LUAF2i(glBindBufferARB)
LUAFglh(glDeleteObjectARB)
LUAFuip(glGenBuffersARB)

LUAFstart(glBufferDataARB)
	int target = LUAPi(1);
	int size = LUAPi(2);
	float *buffer = NULL;
	bool dontfreethebuffer = false;	//free it if its allocated
	int usage = LUAPi(4);
	
	if (!lua_isnil(L,3)) {
		int t = 3;
		if (!lua_istable(L,t)) luaL_error(L, "expected nil or a table as the 3rd param");
		//fill our buffer beforehand
		static vector<float> vf;//not mt safe
		vf.resize(0);
		
		lua_pushnil(L);  // first key
		while (lua_next(L, t) != 0) {
			// 'key' is at index -2 and 'value' at index -1
			vf.push_back((float)lua_tonumber(L,-1));
			lua_pop(L,1);
		}
		
		buffer = &(vf[0]);	//i hope vf sticks around
		dontfreethebuffer = true;
	}
	
#if 0	//debug - init fill somethig fixed ... like 0xdeadbeef
	buffer = (float*)(new char[size]);
	assert(buffer);
	for (int i = 0; i < size/sizeof(float); i++) {
		buffer[i] = 0.f;
		//if ((i&3) == 0 || (i&3) == 3) 
		{
			buffer[i] = 1.f;
		}
		fprintf(stderr, "writing %d is %f\n", i, buffer[i]);
	}
#endif
	glBufferDataARB(target, size, buffer, usage);
	if (buffer && !dontfreethebuffer) delete[] buffer;

//now when i turn this off the whole thing crashes
//why is it so dependent on me mapping and unmapping the buffer?
#if 1
	//try to read back that data and see what you get...
	float *ptr = (float *)glMapBufferARB(target, GL_READ_ONLY);
//	for (int i = 0; i < size/sizeof(float); i++) {
//		fprintf(stderr, "reading %d is %f\n", i, ptr[i]);
//	}
	glUnmapBufferARB(target);
#endif
LUAFend()

void luaBindGL_ARB_vertex_buffer_object(lua_State *L) {
	lua_getglobal(L, "gl");
	
	LUA_GLUE_FUNC_SAFE_ARB(glBindBuffer)
//	LUA_GLUEsARB(glDeleteBuffer)
	LUA_GLUE_FUNC_SAFE_ARB(glDeleteObject)
	LUA_GLUE_FUNC_SAFE_sARB(glGenBuffer)	//im gonna be rendering to this soon ...
//	LUA_GLUEARB(glIsBuffer)
	LUA_GLUE_FUNC_SAFE_ARB(glBufferData)
//	LUA_GLUEARB(glBufferSubData)
//	LUA_GLUEARB(glGetBufferSubData)
//	LUA_GLUEARB(glMapBuffer)
//	LUA_GLUEARB(glUnmapBuffer)
//	LUA_GLUEivARB(glGetBufferParameter)
//	LUA_GLUEvARB(glGetBufferPointer)

/*
  void BindBufferARB(enum target, uint buffer);
    void DeleteBuffersARB(sizei n, const uint *buffers);
    void GenBuffersARB(sizei n, uint *buffers);
    boolean IsBufferARB(uint buffer);

    void BufferDataARB(enum target, sizeiptrARB size, const void *data,
                       enum usage);
    void BufferSubDataARB(enum target, intptrARB offset, sizeiptrARB size,
                          const void *data);
    void GetBufferSubDataARB(enum target, intptrARB offset,
                             sizeiptrARB size, void *data);

    void *MapBufferARB(enum target, enum access);
    boolean UnmapBufferARB(enum target);

    void GetBufferParameterivARB(enum target, enum pname, int *params);
    void GetBufferPointervARB(enum target, enum pname, void **params);
*/

	LUAGiARB(GL_PIXEL_PACK_BUFFER)
	LUAGiARB(GL_PIXEL_PACK_BUFFER)
	LUAGiARB(GL_DYNAMIC_DRAW)
	lua_pop(L,1);
}
