#include <iostream>
#include <map>

#include "system/gl.h"
#include "system/glu.h"
#include "system/exception.h"

#include "base/message.h"	//im sure stl has something like this too
#include "base/image.h"

#include "lib/objfile.h"

#include "program.h"
#include "bindings_util.h"

using namespace std;

//TODO - rename to 'readTex2D' ... better yet, introduce an image subsystem w/bindings
LUAFstart(createTex2D)
	int minFilter = 0;
	int magFilter = 0;
	int unit = 0;
	if (argc < 1) luaL_error(L, "expected at least 1 param");

	const char *filename = NULL;
	if (lua_isstring(L,1)) {
		filename = lua_tostring(L,1);
	} else if (lua_istable(L,1)) {
		lua_getfield(L,1,"file");
		if (lua_isnil(L,-1)) luaL_error(L, "couldnt find the file parameter");
		filename = lua_tostring(L,-1);
		
		lua_getfield(L,1,"minFilter");
		if (!lua_isnil(L,-1)) minFilter = (int)lua_tonumber(L,-1);
		
		lua_getfield(L,1,"magFilter");
		if (!lua_isnil(L,-1)) magFilter = (int)lua_tonumber(L,-1);
		
		lua_getfield(L,1,"unit");
		if (!lua_isnil(L,-1)) unit = (int)lua_tonumber(L,-1);
	}
	if (!filename) luaL_error(L, "couldnt find the filename");

	Image *img = imageSystem.load(filename);
	if (!img) luaL_error(L, "couldnt load the image file %s", filename);

	assert(img->data);
	int bytesPP = img->bitsPerPixel >> 3;	//all we have is 24 or 32 bitsPerPixel atm
	assert(!(img->bitsPerPixel & 7));		//no <8 bit images
	
	GLuint texID;
	glGenTextures(1, &texID);
	const int target = GL_TEXTURE_2D;
	glActiveTextureARB(GL_TEXTURE0 + unit);
	glBindTexture(target, texID);
	gluBuild2DMipmaps(
		target,
		bytesPP == 4 ? 4 : 3,
		VEC2ELEM(img->size),			//auto-adjusts up a power of two, does the proxy testing and everything
		bytesPP == 4 ? GL_RGBA : GL_RGB,
		GL_UNSIGNED_BYTE,
		img->data);

	if (minFilter) glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, minFilter);
	if (magFilter) glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, magFilter);
	glActiveTextureARB(GL_TEXTURE0);

	vec2i size = img->size;

	//no other defaults plz
	delete img;
	
	//sure wthey give them the sizes too
	//in the end i'm just gonna make a table returned, there could be so much info ...
	lua_pushinteger(L, texID);
	lua_pushinteger(L, size.x);
	lua_pushinteger(L, size.y);
	return 3;
LUAFend()

LUAFstart(writeTex2D)
	FILE *file = NULL;
	char *data = NULL;

	if (argc < 1) luaL_error(L, "didnt find a texture index to write out");
	int index = (int)lua_tonumber(L,1);

	//get width & height
	_vec<2,GLint> size;
	glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_WIDTH, &size.x);
	glGetTexLevelParameteriv(GL_TEXTURE_2D, 0, GL_TEXTURE_HEIGHT, &size.y);
	if (size.x <= 0 || size.y <= 0) luaL_error(L, "got a bad size: %d %d", size.x, size.y);

	//then get rgb data
	int pixelsize = 3 * size.x * size.y;
	data = new char[pixelsize];
	if (!data) luaL_error(L, "failed to alloc %d bytes", pixelsize);
	glGetTexImage(GL_TEXTURE_2D, 0, GL_RGB, GL_UNSIGNED_BYTE, data);
	//dont forget to swap rgb with bgr
	for (char *ch = data; ch < data + pixelsize; ch += 3) {
		char tmp = ch[0];
		ch[0] = ch[2];
		ch[2] = tmp;
	}
		
	//try to write it
	const char *fn = "write.bmp";
	file = fopen(fn, "wb");
	if (!file) {
		delete[] data;
		luaL_error(L, "failed to open the file %s for binary writing", fn);
	}

	const int headerSize = 54;
	Message<54> m;
	//bitmap header
	m << (short)0x4d42;
	m << (int)(headerSize + pixelsize);
	m << (int)0;
	m << (int)headerSize;
	assert(m.size == 14);
	//v3 header
	m << (int)40;	//v3 header size
	m << (int)size.x;	//image width
	m << (int)size.y;	//image height
	m << (short)1;	//planes used
	m << (short)24;	//bits per pixel
	m << (int)0;	//compression
	m << pixelsize;	//image size
	m << (int)size.x;	//width ppm
	m << (int)size.y;	//height ppm
	m << (int)0;	//colors used
	m << (int)0;	//important colors used
	assert(m.size == 54);	//depends on message implementation
		
	fwrite(m.data, m.size, 1, file);
	fwrite(data, pixelsize, 1, file);
		
	delete[] data;
	fclose(file);
LUAFend()

LUAFstart(createNormalMap)
	int minFilter = 0;
	int magFilter = 0;
	int unit = 0;
	float relative_altitude = 1;

	if (argc < 1) luaL_error(L, "expected at least 1 param");

	const char *filename = NULL;
	if (lua_isstring(L,1)) {
	
		filename = lua_tostring(L,1);
		if (argc > 1) relative_altitude = lua_tonumber(L,2);
		
	} else if (lua_istable(L,1)) {
		lua_getfield(L,1,"file");
		if (lua_isnil(L,-1)) luaL_error(L, "couldnt find the file parameter");
		filename = lua_tostring(L,-1);
		
		lua_getfield(L,1,"minFilter");
		if (!lua_isnil(L,-1)) minFilter = (int)lua_tonumber(L,-1);
		
		lua_getfield(L,1,"magFilter");
		if (!lua_isnil(L,-1)) magFilter = (int)lua_tonumber(L,-1);
		
		lua_getfield(L,1,"height");
		if (!lua_isnil(L,-1)) relative_altitude = lua_tonumber(L,-1);
		
		lua_getfield(L,1,"unit");
		if (!lua_isnil(L,-1)) unit = (int)lua_tonumber(L,-1);
	}
	if (!filename) luaL_error(L, "couldnt find the filename");

	float inv_alt = 2.f / relative_altitude;
	
	Image *img = imageSystem.load(filename);
	if (!img) luaL_error(L, "unable to load the image file %s", filename);

	assert(img->data);
	int bytesPP = img->bitsPerPixel >> 3;	//all we have is 24 or 32 bitsPerPixel atm
	assert(!(img->bitsPerPixel & 7));		//no <8 bit images

	//make a tangentSpace mpa
	unsigned char *tangentSpace = new unsigned char[img->size.x * img->size.y * 4];
	typedef _quat<unsigned char> quatUC;
	quatUC *p = (quatUC *)tangentSpace;
	for (int j = 0; j < img->size.y; j++) {
		for (int i = 0; i < img->size.x; i++, p++) {

			int im = (i + img->size.x-1) % img->size.x;
			int in = (i + 1) % img->size.x;
			int jm = (j + img->size.y-1) % img->size.y;
			int jn = (j + 1) % img->size.y;

			static vec3f greyscale(0.3f, 0.59f, 0.11f);
			float fim = dot(greyscale, (vec3f)*(vec3uc*)&img->data[bytesPP * (im + j * img->size.x)]) / 255.f;
			float fin = dot(greyscale, (vec3f)*(vec3uc*)&img->data[bytesPP * (in + j * img->size.x)]) / 255.f;
			float fjm = dot(greyscale, (vec3f)*(vec3uc*)&img->data[bytesPP * (i + jm * img->size.x)]) / 255.f;
			float fjn = dot(greyscale, (vec3f)*(vec3uc*)&img->data[bytesPP * (i + jn * img->size.x)]) / 255.f;
			//surface normal is <-dx, -dy, 2/z_scale>
			//computation of <2, 0, z * dx> x <0, 2, z * dy>
			//does ~ take precedence over + ?
			//vec3f del = ( vec3f(fim - fin, fjm - fjn, inv_alt).normalize() + 1.f) * 0.5f;
			//all in place:
			vec3f del(fim-fin, fjm-fjn, inv_alt);
			(del.normalize() += 1.f) *= 0.5f;

			p->axis() = (vec3uc)(del * 255.f);
			//store the actual height:
			p->w = (unsigned char)dot(greyscale, (vec3f)*(vec3uc*)&img->data[bytesPP * (i + j * img->size.x)] );
		}
	}
		
	GLuint texID;
	glGenTextures(1, &texID);
	const int target = GL_TEXTURE_2D;
	glActiveTextureARB(GL_TEXTURE0 + unit);
	glBindTexture(target, texID);
	gluBuild2DMipmaps(target, 4, VEC2ELEM(img->size), GL_RGBA, GL_UNSIGNED_BYTE, tangentSpace);
	glActiveTextureARB(GL_TEXTURE0);

	//no other defaults plz
	delete[] tangentSpace;
	delete img;
		
	lua_pushinteger(L, texID);
	return 1;
LUAFend()

LUAFstart(createTexCubeMap)
	if (argc != 6) luaL_error(L, "expected six filenames");
		
	GLuint cubeMapTex;
	glGenTextures(1, &cubeMapTex);
	glBindTexture(GL_TEXTURE_CUBE_MAP, cubeMapTex);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP);
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP);
	// turn on auto mipmapping
	glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_GENERATE_MIPMAP, true);	// will lua correctly convert 'true' to '1' ?
	glHint(GL_GENERATE_MIPMAP_HINT, GL_NICEST);	// GL_NICEST or GL_FASTEST
	// make the individual face textures (empty)
	for (int i = 0; i < 6; i++) {
	
		const char *filename = lua_tostring(L,i+1);
		if (!filename) luaL_error(L, "couldnt find the filename %d", (i+1));

		Image *img = imageSystem.load(filename);
		if (!img) luaL_error(L, "couldnt load the image file %s", filename);

		assert(img->data);
		int bytesPP = img->bitsPerPixel >> 3;	//all we have is 24 or 32 bitsPerPixel atm
		assert(!(img->bitsPerPixel & 7));		//no <8 bit images

		int target = GL_TEXTURE_CUBE_MAP_POSITIVE_X + i;
//		glTexImage2D(target, 0, bytesPP, VEC2ELEM(img->size), 0, bytesPP == 4 ? GL_RGBA : GL_RGB, GL_UNSIGNED_BYTE, img->data);
		gluBuild2DMipmaps(target, bytesPP, VEC2ELEM(img->size), bytesPP == 4 ? GL_RGBA : GL_RGB, GL_UNSIGNED_BYTE, img->data);
	}
		
	lua_pushinteger(L, cubeMapTex);
	return 1;
LUAFend()

LUAFstart(createTex3D)
	GLuint texid = 0;
	char *data = NULL;
	FILE *file = NULL;

	if (argc != 1) luaL_error(L, "expected a filename");
	const char *fn = lua_tostring(L,1);
	if (!fn) luaL_error(L, "couldnt find the filename");
		
	file = fopen(fn, "rb");
	if (!file) luaL_error(L, "failed to open file %s", fn);

	vec3i size;
	fread(&size.x, 4, 1, file);
	fread(&size.y, 4, 1, file);
	fread(&size.z, 4, 1, file);
	if (size.x <= 0 || size.y <= 0 || size.z <= 0) {
		fclose(file);
		luaL_error(L, "got bad size %d %d %d ", VEC3ELEM(size));
	}
		
	int filesize = size.x*size.y*size.z*4;
	data = new char[filesize];	//rgba
	if (!data) {
		fclose(file);
		luaL_error(L, "failed to alloc buffer of size %s", filesize);
	}
		
	int r = fread(data, 1, filesize, file);
	if (r != filesize) luaL_error(L, "only read %d of %d bytes", r, filesize);
		
	//now do our texture junk
	glGenTextures(1, &texid);
	glBindTexture(GL_TEXTURE_3D, texid);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
#ifdef WIN32
	glTexImage3DEXT
#else
	glTexImage3D
#endif
		(GL_TEXTURE_3D, 0, GL_RGBA, VEC3ELEM(size), 0, GL_RGBA, GL_UNSIGNED_BYTE, data);

	if (file) fclose(file);
	if (data) delete data;
	
	if (texid) {
		lua_pushinteger(L, texid);
		return 1;
	}
LUAFend()


class ObjWrapper {
public:
	ObjFile *obj;
	
	class AttribVector {
	public:
		string name;
		int dim;
		vector<float> elems;	//m*n
	};
	
	//we want an array of floats or something for our object ...
	//how will that work?
	//vector of floats sized m*n (m is # of vertices, n is # of elements)
	vector< AttribVector > attribs;

	ObjWrapper(ObjFile *_obj) : obj(_obj) {}
	
	void addAttrib(const string &name, int dim) {
		int index = attribs.size();
		attribs.resize( index+1 );
		attribs[index].name = name;
		attribs[index].dim = dim;
		attribs[index].elems.resize(dim * obj->getVertexes().size());
	}
};

//the vector is for index-based obj files
static vector<ObjWrapper *> objFiles;
//the map is for string-based ones
static map<string, int> objFileMap;		//map filenames to locations within the obj vector

static int doLoadObj(const char *fn, int loadFlags = 0) {
	if (!fn) {
		con << "doLoadObj got a null filename" << endl;
		return -1;
	}

	///look for the string in the map (prevent multiple loading)
	string sfn(fn);
	map<string, int>::iterator i = objFileMap.find(sfn);
	if (i != objFileMap.end()) {
		return i->second;
	}

	ObjFile *file = new ObjFile();
	if (!file) return -1;

	if (!file->load(fn, loadFlags)) {	//no bitflags
		delete file;
		con << "doLoadOBJ failed to load file " << fn << endl;
		return -1;
	}
	file->compile();

	//put the file in the list
	objFiles.push_back(new ObjWrapper(file));
	//get the list index
	int index = (int)objFiles.size()-1;
	//and add it to the map
	objFileMap[string(fn)] = index;

	return index;
}

LUAFstart(loadOBJ)
	if (!lua_isstring(L,1)) {
		con << "loadOBJ expected a string as param 1" << endl;
	} else {
		int loadflags = 0;
		if (lua_isstring(L,2) && !strcmp(lua_tostring(L,2),"regenNormals")) {
con << "regen'ing normals" << endl;
			loadflags |= bitflag(ObjFile::REGEN_NORMALS);
		}
		int index = doLoadObj(lua_tostring(L,1), loadflags);
		if (index == -1) return 0;
		lua_pushinteger(L,index);
		return 1;
	}
LUAFend()

#define LUAFOBJstart(x)	\
LUAFstart(x)	\
	if (argc < 1) {	\
		con << #x " expected an index number" << endl;	\
		return 0;	\
	} 	\
	int index = -1;	\
	if (lua_isnumber(L,1)) {	\
		index = (int)lua_tonumber(L,1);	\
	} else if (lua_isstring(L,1)) {	\
		index = doLoadObj(lua_tostring(L,1));	\
		if (index == -1) return 0;	\
	} else {	\
		con << #x " got a bad type" << endl;	\
		return 0;	\
	}	\
	if (index < 0 || index >= objFiles.size()) {	\
		con << #x " got a bad index: " << index	\
			<< " should be between 0 and " << (objFiles.size()-1);	\
	}	\
	ObjWrapper *obj = objFiles[index];
	
LUAFOBJstart(objCreateAttributeArray)
	string name = lua_tostring(L,2);
	int dim = (int)lua_tonumber(L,3);
	obj->addAttrib(name, dim);
LUAFend()

LUAFOBJstart(objSetAttributeArray)
	string name = lua_tostring(L,2);
	for (int i = 0; i < obj->attribs.size(); i++) {
		ObjWrapper::AttribVector &attr = obj->attribs[i];
		if (attr.name == name) {
			int index = (int)lua_tonumber(L,3);
			index *= attr.dim;
			for (int j = 0; j < attr.dim; j++) {
				attr.elems[index+j] = lua_tonumber(L,4+j);
			}
			return 0;
		}
	}
	con << "objSetAttributeArray couldnt find teh attrib named " << name << endl;
LUAFend()

LUAFOBJstart(drawOBJ)
	obj->obj->render();
	lua_pushinteger(L, index);
	return 1;
LUAFend()

LUAFOBJstart(objDrawWithArrays)
	obj->obj->renderWithArrays();
	lua_pushinteger(L, index);
	return 1;
LUAFend()

LUAFOBJstart(objVertexCount)
	lua_pushinteger(L, obj->obj->getVertexes().size());
	return 1;
LUAFend()

LUAFOBJstart(objNormalCount)
	lua_pushinteger(L, obj->obj->getNormals().size());
	return 1;
LUAFend()

LUAFOBJstart(objTriCount)
	lua_pushinteger(L, obj->obj->getTris().size());
	return 1;
LUAFend()

LUAFOBJstart(objTriVertex)
	int triindex = (int)lua_tonumber(L,2);
	const vector<tri_t> &tris = obj->obj->getTris();
	if (triindex < 0 || triindex >= tris.size()) {
		con << "objTriVertex object " << index << " got oob tri index " << triindex << " of 0 to " << tris.size() << endl;
		return 0;
	}
	int vtxsubindex = (int)lua_tonumber(L,3);
	if (vtxsubindex < 0 || vtxsubindex > 2) {
		con << "objTriVertex object " << index << " got oob vertex index " << vtxsubindex << " of 0 to 2" << endl;
	}
	int vtxindex = tris[triindex].p[vtxsubindex].v;
	const vector<vec3f> &vtxs = obj->obj->getVertexes();
	if (vtxindex < 0 || vtxindex >= vtxs.size()) {
		con << "objTriVertex got oob index " << vtxindex << endl;
		return 0;
	}
	const vec3f &v = vtxs[vtxindex];
	for (int i = 0; i < 3; i++) {
		lua_pushnumber(L, v[i]);
	}
	return 3;
LUAFend()

LUAFOBJstart(objTriVtxIndex)
	int triindex = (int)lua_tonumber(L,2);
	const vector<tri_t> &tris = obj->obj->getTris();
	if (triindex < 0 || triindex >= tris.size()) {
		con << "objTriVertex object " << index << " got oob tri index " << triindex << " of 0 to " << tris.size() << endl;
		return 0;
	}
	int vtxsubindex = (int)lua_tonumber(L,3);
	if (vtxsubindex < 0 || vtxsubindex > 2) {
		con << "objTriVertex object " << index << " got oob vertex index " << vtxsubindex << " of 0 to 2" << endl;
	}
	int vtxindex = tris[triindex].p[vtxsubindex].v;
	const vector<vec3f> &vtxs = obj->obj->getVertexes();
	if (vtxindex < 0 || vtxindex >= vtxs.size()) {
		con << "objTriVertex got oob index " << vtxindex << endl;
		return 0;
	}
	lua_pushinteger(L, vtxindex);
	return 1;
LUAFend()

LUAFOBJstart(objTriNormal)
	int triindex = (int)lua_tonumber(L,2);
	const vector<tri_t> &tris = obj->obj->getTris();
	if (triindex < 0 || triindex >= tris.size()) {
		con << "objTriNormal object " << index << " got oob tri index " << triindex << " of 0 to " << tris.size() << endl;
		return 0;
	}
	int vtxsubindex = (int)lua_tonumber(L,3);
	if (vtxsubindex < 0 || vtxsubindex > 2) {
		con << "objTriNormal object " << index << " got oob vertex index " << vtxsubindex << " of 0 to 2" << endl;
		return 0;
	}
	int vtxindex = tris[triindex].p[vtxsubindex].n;
	const vector<vec3f> &vtxs = obj->obj->getNormals();
	if (vtxindex < 0 || vtxindex >= vtxs.size()) {
		con << "objTriNormal got oob index " << vtxindex << endl;
		return 0;
	}
	const vec3f &v = vtxs[vtxindex];
	for (int i = 0; i < 3; i++) {
		lua_pushnumber(L, v[i]);
	}
	return 3;
LUAFend()

//takes in a obj id and a vertex index, returns the position
LUAFOBJstart(objVertex)
	int vtxindex = (int)lua_tonumber(L,2);
	const vector<vec3f> &vtxs = obj->obj->getVertexes();
	if (vtxindex < 0 || vtxindex >= vtxs.size()) {
		con << "objVertex object " << index << " got oob vtx index " << vtxindex << " of 0 to " << vtxs.size() << endl;
		return 0;
	}
	const vec3f &v = vtxs[vtxindex];
	for (int i = 0; i < 3; i++) {
		lua_pushnumber(L, v[i]);
	}
	return 3;
LUAFend()

LUAFOBJstart(objNormal)
	int normalindex = (int)lua_tonumber(L,2);
	const vector<vec3f> &normals = obj->obj->getNormals();
	if (normalindex < 0 || normalindex >= normals.size()) {
		con << "objVertex object " << index << " got oob normal index " << normalindex << " of 0 to " << normals.size() << endl;
		return 0;
	}
	const vec3f &v = normals[normalindex];
	for (int i = 0; i < 3; i++) {
		lua_pushnumber(L, v[i]);
	}
	return 3;
LUAFend()

static GLUquadric *initQuadric() {
	static GLUquadric *quadric = NULL;
	if (!quadric) {
		quadric = gluNewQuadric();
		gluQuadricNormals(quadric, GLU_SMOOTH);
		gluQuadricTexture(quadric, true);
	}
	return quadric;
}

LUAFstart(drawCylinder)
	GLUquadric *q = initQuadric();
	double base = 1;
	double top = 1;
	double height = 1;
	int slices = 1;
	int stacks = 1;
	gluCylinder(q, base, top, height, slices, stacks);
LUAFend()

LUAFstart(drawSphere)
	GLUquadric *q = initQuadric();
	double radius = 1;
	int slices = 20;
	int stacks = 20;
	gluSphere(q, radius, slices, stacks);
LUAFend()

LUAFstart(vectorToAngleAxis)
	vec3f v1(0,0,-1);
	vec3f v2 = luaReadVec3f(L).normalize();	//(LUAPd(1),LUAPd(2),LUAPd(3));
	vec4f q;
	if (dot(v1,v2) < -.9999f) {
		q.set(0,1,0,180);
	} else {
		q = quatToAngleAxis(quatVectorDifference(v1, v2, 1.f));
	}
	lua_pushnumber(L, q.x);
	lua_pushnumber(L, q.y);
	lua_pushnumber(L, q.z);
	lua_pushnumber(L, q.w);
	return 4;
LUAFend()

LUAFstart(gluPickMatrix)
	GLint vp[4] = {
		lua_tointeger(L, 5),
		lua_tointeger(L, 6),
		lua_tointeger(L, 7),
		lua_tointeger(L, 8),
	};
	gluPickMatrix(
		lua_tonumber(L, 1),
		lua_tonumber(L, 2),
		lua_tonumber(L, 3),
		lua_tonumber(L, 4),
		vp);
LUAFend()

/*
GLint
gluUnProject(GLdouble winx, GLdouble winy, GLdouble winz,
		const GLdouble modelMatrix[16], 
		const GLdouble projMatrix[16],
                const GLint viewport[4],
	        GLdouble *objx, GLdouble *objy, GLdouble *objz)
			
... becomes ...

result, objx, objy, objz = gluUnProject(win, modelMatrix, projMatrix, viewport)
*/
LUAFstart(gluUnProject)
	vec3d win;
	for (int i = 0; i < 3; i++) {
		lua_pushinteger(L, i+1);
		lua_gettable(L, 1);
		win[i] = lua_tonumber(L, -1);
		lua_pop(L, 1);
	}
	
	double modelMatrix[16];
	for (int i = 0; i < 16; i++) {
		lua_pushinteger(L, i+1);
		lua_gettable(L, 2);
		modelMatrix[i] = lua_tonumber(L, -1);
		lua_pop(L, 1);
	}

	double projMatrix[16];
	for (int i = 0; i < 16; i++) {
		lua_pushinteger(L, i+1);
		lua_gettable(L, 3);
		projMatrix[i] = lua_tonumber(L, -1);
		lua_pop(L, 1);
	}
	
	int viewport[4];
	for (int i = 0; i < 4; i++) {
		lua_pushinteger(L, i+1);
		lua_gettable(L, 4);
		viewport[i] = lua_tointeger(L, -1);
		lua_pop(L, 1);
	}
	
	vec3d obj;
	int res = gluUnProject(
		VEC3ELEM(win),
		modelMatrix,
		projMatrix,
		viewport,
		&obj.x, &obj.y, &obj.z);
		
	lua_pushinteger(L, res);
	lua_pushnumber(L, obj.x);
	lua_pushnumber(L, obj.y);
	lua_pushnumber(L, obj.z);
	return 4;
LUAFend()

static const luaL_reg utilBindings[] = {
//texture
	LUA_GLUE(createTex2D)	//load a 2d tex
	LUA_GLUE(writeTex2D)
	LUA_GLUE(createNormalMap)	//load a 2d tex, convert it to greyscale, pop out a rgb normal map w/alpha the height component
	LUA_GLUE(createTexCubeMap)	//load six 2d texs as cubemaps
	LUA_GLUE(createTex3D)
	LUA_GLUE(gluPickMatrix)
	LUA_GLUE(gluUnProject)
//shader
//	LUA_GLUE(createShader)
//wavefront 3d object file
	LUA_GLUE(loadOBJ)	// or i could return a table with a metatable set to all the obj routines ...
	LUA_GLUE(drawOBJ)
	LUA_GLUE(objDrawWithArrays)
	LUA_GLUE(objVertexCount)
	LUA_GLUE(objNormalCount)
	LUA_GLUE(objVertex)
	LUA_GLUE(objNormal)
	LUA_GLUE(objTriCount)
	LUA_GLUE(objTriVertex)
	LUA_GLUE(objTriVtxIndex)
	LUA_GLUE(objTriNormal)
	LUA_GLUE(objCreateAttributeArray)
	LUA_GLUE(objSetAttributeArray)
//glu quadrics
	LUA_GLUE(drawCylinder)
	LUA_GLUE(drawSphere)
//math.  its actually faster.  i think everything is.  gosh.
	LUA_GLUE(vectorToAngleAxis)
//terminator
	{NULL, NULL}
};

void luaBind_Util(lua_State *L) {
	lua_getglobal(L, "glush");
	luaAddBindings(L, utilBindings);
	lua_pop(L,1);
}
