/*

    File: libglush.cpp

    Copyright (C) 2000-2010 Christopher Moore (christopher.e.moore@gmail.com)
	  
    This software is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
  
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License along
    with this program; if not, write the Free Software Foundation, Inc., 51
    Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/
#include <iostream>

#include <lua.hpp>

#include "main.h"
#include "program.h"
#include "bindings.h"

using namespace std;

static int libglush_run(lua_State *L) {
	int args = lua_gettop(L);
	if (args >= 1) {
		if (lua_isfunction(L, 1)) {
			lua_getglobal(L, "glush");			//display, ..., glush
			lua_insert(L, 1);					//display, ..., glush, display
			lua_setfield(L, lua_gettop(L)-2, "display");	//display, ..., glush
			lua_pop(L,1);					//display
		}
	}
	
	//here's the kicker:
	//glushModule is a module
	//that runs sys
	//that creates a carbon window
	//	that's not what we want
	//	we want to pop up a new application
	//	like virtualbox does

	//this doesn't work so well
	//it pops up the carbon window behind the terminal
	//and certainly doesn't capture the keyboard at all
	//seems a better way would be to launch an encapsulated application?
	//or just keep it an engine?

	// set up the module with the callbacks here.
	glushModule.loop();
	
	//since there is no luaclose_glush, lets just shut down here
	//and only let this run once...?
	//or let the OS handle shutdown and let it run as much as it wants?
	//nah, that causes crashes...
	glushModule.shutdown();
	
	//somewhere after here we're getting a double free error
	//probably a global's dtor
	return 0;
}

void sysBindSingletons();

extern "C" {
#ifdef WIN32
__declspec(dllexport)
#endif
int luaopen_glush(lua_State *L) {

	//TODO - it'd sure be nice to not bind singletons, lua state, or anything
	//until "glush.run()" was called

	//call init as soon as glush is required
	//otherwise any subsequent gl calls will have no context and crash
	
	//this could be circumvented by forcing all scripts to use a dedicated "glush.init" aspect...
	//....nah...
	sysBindSingletons();

	glushModule.initWithGoods(L);
	
	lua_getglobal(L, "glush");
	assert(lua_istable(L, -1));
	lua_pushcfunction(L, libglush_run);
	lua_setfield(L, -2, "run");

	luaStateAddMainCalls(L);
	luaBind(L);	//any additional bindings?
	glushMenu.bind(L);

	return 0;
}
}
